/* CSVToXMLBasic.cpp

This program converts an input comma separated
value file to an XML instance document
of a Row/ColumXX organization

Program developed in Chapter 3
Using XML with Legacy Business Applications

Source code is Copyright (c) 2002 Addison-Wesley Professional
Author:
Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License,
Version 2, as published by the Free Software Foundation.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"

//  Projet header files
#include "CSVRowReader.h"


int main(int argc, char* argv[])
{

  //  Local variables
  //  Text as char * constants

  char cUsageMessage[] =
	  "Usage: CSVToXMLBasic InputFile.CSV OutputFile.XML options \
or CSVToXMLBasic -h";

  char cHelpMessage[] = 
"Converts an input comma separated value file  \n\
into an output XML document in Row/ColumnXX format. \n\
\n\
Arguments:\n\
  InputFile.CSV    Input CSV file\n\
  OutputFile.XML   Output XML instance document\n\
\n\
Options:\n\
  -h   Display this message and exit with no action\n\
\n\
Copyright (c) 2002 Addison-Wesley Professional\n\
Author: Michael C. Rawlins, Rawlins EC Consulting\n\
\n\
CSVToXMLBasic\
 comes with with ABSOLUTELY NO WARRANTY.\n\
This is free software and you are welcome to\n\
redistribute it under certain conditions.\n\
Refer to the GNU General Public License for details.\n";

  char cHeaderText[] = "version=\"1.0\" encoding=\"UTF-8\"";

  //  Project Classes
  CSVRowReader * RowReader;

  //  COM
  HRESULT hResult;
  
  //  Input and output streams
  ifstream InputCSV;

  //  Basic types
  char * cInputCSVName, * cOutputXMLName;
  char cInputRecord[MAXCOLUMNSIZE];

  int iExitStatus, iRows;

  //  Constants - Homegrown exception values
  const char cCreateInstanceError = 1;
  const char cSaveError = 2;
  const char cInputFileOpenError = 3;
  const char cInputFileReadError = 4;

  // Initializations
  iExitStatus = 0;
  iRows = 0;
  hResult = S_OK;


  // Parse input arguments from command line
  // IF help option is specified
  //     display help message and exit
  // ENDIF
  if ( ( ( argc == 2 ) && (strcmp(argv[1],"-h") == 0 ) ) ||
       ( ( argc == 4 ) && (strcmp(argv[3],"-h") == 0 ) ) )
  {
    cout << cUsageMessage << endl << endl;
    cout << cHelpMessage << endl;
    return 0;
  }
    
  if ((argc < 3) || (argc > 4) )
  {
    cout << cUsageMessage << endl;
    return 1;
  }

  cInputCSVName = argv[1];
  cOutputXMLName = argv[2];


  //  Set up the COM environment, since MSXML is a
  //  COM Component
  hResult = CoInitialize(NULL);
  if (FAILED(hResult))
  {
    cerr << "Failed to initialize COM environment" << endl;
    return 0;
  }

  //  We'll do our main processing within a try block
  //  We only expect to throw COM, MSXML parse
  //  and I/O exceptions
  try
  {
    //  Declare smart pointers for our MSXML COM Interfaces.
    //  We'll use the Dccument2 class for all of our stuff so
    //  that we can do schema validation.
    //  We do these within the try block so that they when we
    //  exit the program they will be out of scope.
    IXMLDOMDocument2Ptr spDocOutput;
    IXMLDOMElementPtr spEleRoot;
    IXMLDOMProcessingInstructionPtr spPIHeader;

  	// Open input file
	  InputCSV.open(cInputCSVName, ios::in );
  	if ( ! InputCSV )
    {
      throw cInputFileOpenError;
    }

    //  Set up DOM XML environment (dependent on implementation)
    //  Create the Output XML Document (dependent on
    //    implementation)
    //  Create the COM DOM Document object
    hResult = spDocOutput.CreateInstance(
      __uuidof(DOMDocument40));

    if FAILED(hResult)
    {
      throw cCreateInstanceError;
    }

    //  Create the header processing instruction
    //    and append it to the document node
    spPIHeader = spDocOutput->createProcessingInstruction("xml",
      cHeaderText);
    spDocOutput->appendChild(spPIHeader);

    //  Root Element <- Call Document's createElement method,
    //    with tagName of SimpleCSV
    spEleRoot = spDocOutput->createElement("SimpleCSV");

    //  Document <-  Call Document's appendChild to append Root
    //    to Document
    spDocOutput->appendChild(spEleRoot);

    //  Initialize CSVRowReader object using the Output Document
    //    and the Root Element
    RowReader = new CSVRowReader(spDocOutput,spEleRoot);

    //  Read first record from file
    memset(cInputRecord,'\0',MAXCOLUMNSIZE);
    InputCSV.getline(cInputRecord,MAXCOLUMNSIZE);
    if (InputCSV.bad())
    {
      throw cInputFileReadError;
    }
    
    //  DO until all records are read
  	//    Increment Row Count
    //    Call CSVRowReader parse method, passing the record
    //      read from the input file
    //    Call CSVRowReader write method
    //    Read next record from input file
    //  ENDDO
    while (!InputCSV.eof())
    {
      iRows++;
      RowReader->parse(cInputRecord);
      RowReader->write();
      memset(cInputRecord,'\0',MAXCOLUMNSIZE);
      InputCSV.getline(cInputRecord,MAXCOLUMNSIZE);
      if (InputCSV.bad())
      {
        throw cInputFileReadError;
      }
    }

    //  Save Output XML Document (dependent on implementation)
    //  Save it using the document's save method
    hResult = spDocOutput->save(cOutputXMLName);

    if FAILED(hResult)
    {
      throw cSaveError;
    }
 
  } // End of try block

 
  //  General purpose catch block
  catch (char cException)
  {
    switch (cException)
    {
    case  cCreateInstanceError:
      cerr << "Failed to create Document instance" << endl;
      cerr << "HRESULT = x" << hex << hResult << endl;
    case  cSaveError:
      cerr << "Failed to save XML document" << endl;
      break;
    case  cInputFileOpenError:
      cerr << "Failure in opening input file " << 
        cInputCSVName << endl;
      break;
    case  cInputFileReadError:
      cerr << "Error in reading input file " <<
        cInputCSVName << endl;
      break;
    default:
      cerr << "Unexpected char exception " << cException
        << endl;
    }
    iExitStatus = 1;
  }
    
  //  Do some simple error handling for COM exceptions
  catch (_com_error &e)
  {
    handleCOMError(e);
    iExitStatus = 1;
  }

  // Finish up
  if (iExitStatus == 0)
  {
    cout << "Number of CSV rows read: " << iRows << endl;
  }
  else
  {
    cerr << "Abnormal Termination" << endl;
  }

  //  Cleanup
  if (InputCSV != NULL)
  {
    InputCSV.close();
  }

  //  It is advisable to clean up C++ objects before
  //  COM uninitialize
  if (RowReader != NULL)
  {
    delete RowReader;
  }

  CoUninitialize();

  return iExitStatus;
}
