/* X12RecordWriter.cpp

Creates X12 EDI control segments

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

#include <time.h>


//  Common header files
#include "BlasterIncludes.h"

//  Special headers for this module only
#include "EDIRecordWriter.h"
#include "X12RecordWriter.h"

//  Constructor with output stream
X12RecordWriter::X12RecordWriter(
  const IXMLDOMDocument2Ptr spFileDescription,
  ofstream * const OutputFileIn) :
  EDIRecordWriter(spFileDescription, OutputFileIn)
{ }

//  Destructor
X12RecordWriter::~X12RecordWriter() {};


// Write the ISA Interchange Control Header Segment
int X12RecordWriter::writeISA()
{
  //  Locals
  //  XML
  IXMLDOMElementPtr spISAInfo;
  IXMLDOMElementPtr spX12Control;
  IXMLDOMElementPtr spTemp;
  IXMLDOMNodeListPtr spTPNodeList;

  time_t ltime;
  struct tm *today;

  //  Basic
  int iInterchangeControlNumber;
  int i, iLen, iPad;
  char cISASegment[107];
  //  For ISA data elements
  char cISA[17][32];
  char cISAReceiverID[32];
  char cControlReceiverID[32];
  char cYear[5];
  char cMonth[3];
  char cDay[3];
  char cHour[3];
  char cMinute[3];

  iFunctionalGroupCount = 0;

  memset(cControlReceiverID,0,sizeof(cControlReceiverID));
  memset(cISASegment,0,sizeof(cISASegment));
  for(i=0; i < 17; i++)
  {
    memset(cISA[i],0,sizeof(cISA[i]));
  }

  //  ISAInfo <- Call File Description Document's
  //    getElementsByTagName for "ISAInformation"
  spISAInfo =
      spFileDescription->getElementsByTagName(
                                    "ISAInformation")->item[0];

  //  ISAReceiverID <- Call ISAInfo's getElementsByTagName for
  //    "ReceiverID", then getAttribute for "value'
  spTemp = 
      spISAInfo->getElementsByTagName("ReceiverID")->item[0];
  memset(cISAReceiverID,0,sizeof(cISAReceiverID));
  strcpy(cISAReceiverID,
      _bstr_t(spTemp->getAttribute("value")));

  //  X12Control <- Call EDI Control documemnt's
  //    getElementsByTagName for "X12OutboundControl"
  spX12Control = 
      spEDIControl->getElementsByTagName(
                                "X12OutboundControl")->item[0];

  //  TPNodeList <- Call X12Control's getElementsByTagName for
  //    "TradingPartner"
  spTPNodeList =
      spX12Control->getElementsByTagName("TradingPartner");
  iLen = spTPNodeList->length;

  //  DO for all entries in TPNodeList or until
  //    ControlReceiverID = ISAReceiverID
  i = 0;
  while (i < iLen)
  {
    //  ControlReceiverID <- call TPNodeList item's
    //    getAttribute for "ReceiverID"
    memset(cControlReceiverID,0,sizeof(cControlReceiverID));
    IXMLDOMElementPtr spTP = spTPNodeList->item[i];
    strcpy(cControlReceiverID,
        _bstr_t(spTP->getAttribute("ReceiverID")));
    if (!strcmp(cControlReceiverID,cISAReceiverID))
    {
      spTPElement = spTP;
      break;
    }
    i++;
  } //  ENDDO

  //  IF no match found
  if ( i == iLen )
  {
    //  TPElement <- call EDI Control's createElement method
    //    using Element name of "TradingPartner"
    spTPElement =
        spEDIControl->createElement("TradingPartner");
    //  X12Control <- call X12Control's appendChild to attach
    //    TPElement
    spX12Control->appendChild(spTPElement);
    iInterchangeControlNumber = 1;
    //  Call TPElement's setAttribute for "ReceiverID", passing
    //    ISAReceiverID as value to set
    spTPElement->setAttribute("ReceiverID", cISAReceiverID);
  }
  else
  {
    //  Interchange Number <- Call TP Element's getAttribute
    //    for "LastInterchangeNumber"
    memset(cInterchangeControlNumber,0,
        sizeof(cInterchangeControlNumber));
    strcpy(cInterchangeControlNumber,
        _bstr_t(
          spTPElement->getAttribute("LastInterchangeNumber")));
    iInterchangeControlNumber = atoi(cInterchangeControlNumber);
    iInterchangeControlNumber++;
  } //  ENDIF

  //  Call TPElement's setAttribute for "LastInterchangeNumber"
  //    passing Interchange Number as value to set
  _itoa(iInterchangeControlNumber, cInterchangeControlNumber, 10);
  spTPElement->setAttribute("LastInterchangeNumber",
                            cInterchangeControlNumber);

  //  Date <- Get date from system, and reformat to YYMMDD
  //  Time <- Get time from system, and reformat to HHMM
  // Set time zone from TZ environment variable. If TZ is not set,
  // the operating system is queried to obtain the default value 
  // for the variable. 
  //
  tzset();

  // Get System Time
  time( &ltime );
  today = localtime( &ltime );
  memset(cYear,0,sizeof(cYear));
  memset(cMonth,0,sizeof(cMonth));
  memset(cDay,0,sizeof(cDay));
  memset(cHour,0,sizeof(cHour));
  memset(cMinute,0,sizeof(cMinute));

  sprintf(cYear,"%d",(today->tm_year + 1900));
  sprintf(cMonth,"%02d",(today->tm_mon + 1));
  sprintf(cDay,"%02d",(today->tm_mday ));
  sprintf(cHour,"%02d",(today->tm_hour ));
  sprintf(cMinute,"%02d",(today->tm_min ));
  memset(cDate8,0,sizeof(cDate8));
  memset(cDate6,0,sizeof(cDate6));
  memset(cTime,0,sizeof(cTime));
  strcpy(cDate8,cYear);
  strcat(cDate8,cMonth);
  strcat(cDate8,cDay);
  strcpy(cDate6,&(cDate8[2]));
  
  strcpy(cTime,cHour);
  strcat(cTime,cMinute);

  //  Get remaining values from ISAInfo Element and pad to
  //    required Lengths
  spTemp = spISAInfo->getElementsByTagName(
                            "AuthorizationQualifier")->item[0];
  strcpy(cISA[1], _bstr_t(spTemp->getAttribute("value")));

  spTemp = spISAInfo->getElementsByTagName(
                                "AuthorizationInfo")->item[0];
  strcpy(cISA[2], _bstr_t(spTemp->getAttribute("value")));

  spTemp = spISAInfo->getElementsByTagName(
                                "SecurityQualifier")->item[0];
  strcpy(cISA[3], _bstr_t(spTemp->getAttribute("value")));

  spTemp = 
    spISAInfo->getElementsByTagName("SecurityInfo")->item[0];
  strcpy(cISA[4], _bstr_t(spTemp->getAttribute("value")));

  spTemp = spISAInfo->getElementsByTagName(
                                  "SenderIDQualifier")->item[0];
  strcpy(cISA[5], _bstr_t(spTemp->getAttribute("value")));

  spTemp = spISAInfo->getElementsByTagName("SenderID")->item[0];
  strcpy(cISA[6], _bstr_t(spTemp->getAttribute("value")));

  spTemp = spISAInfo->getElementsByTagName(
                              "ReceiverIDQualifier")->item[0];
  strcpy(cISA[7], _bstr_t(spTemp->getAttribute("value")));

  strcpy(cISA[8],cISAReceiverID);
  strcpy(cISA[9],cDate6);
  strcpy(cISA[10],cTime);

  //  Support for Repetition Separator introduced in 004020
  if (cRepetitionSeparator != 0)
  {
    cISA[11][0] = cRepetitionSeparator;
  }
  else
  {
    cISA[11][0] = 'U';
  }

  spTemp =
    spISAInfo->getElementsByTagName("VersionNumber")->item[0];
  strcpy(cISA[12], _bstr_t(spTemp->getAttribute("value")));

  strcpy(cISA[13],cInterchangeControlNumber);

  //  We hard code a zero for Interchange Acknowledgement
  //    Requested - TA1 segments are not normally used
  cISA[14][0] = '0';

  spTemp = 
    spISAInfo->getElementsByTagName("TestIndicator")->item[0];
  strcpy(cISA[15], _bstr_t(spTemp->getAttribute("value")));

  cISA[16][0] = cComponentSeparator;

  //  Pad as need
  if (strlen(cISA[1]) < 2)
  {
    cISA[1][1] = ' ';
  }

  iLen = strlen(cISA[2]);
  if (iLen < 10)
  {
    iPad = 10 - iLen;
    memset(&(cISA[2][iLen]),' ',iPad);
  }

  if (strlen(cISA[3]) < 2)
  {
    cISA[3][1]= ' ';
  }

  iLen = strlen(cISA[4]);
  if (iLen < 10)
  {
    iPad = 10 - iLen;
    memset(&(cISA[4][iLen]),' ',iPad);
  }

  iLen = strlen(cISA[6]);
  if (iLen < 15)
  {
    iPad = 15 - iLen;
    memset(&(cISA[6][iLen]),' ',iPad);
  }

  iLen = strlen(cISA[8]);
  if (iLen < 15)
  {
    iPad = 15 - iLen;
    memset(&(cISA[8][iLen]),' ',iPad);
  }

  iLen = strlen(cISA[13]);
  if (iLen < 9)
  {
    iPad = 9 - iLen;
    strcpy(&(cISA[13][iPad]),cISA[13]);
    memset(cISA[13],'0',iPad);
  }

  //  Save the full length Interchange Control Number
  memset(cInterchangeControlNumber,0,
      sizeof(cInterchangeControlNumber));
  strcpy(cInterchangeControlNumber, cISA[13]);

  //  Build the ISA segment, starting with the tag
  memset(cISASegment,0,sizeof(cISASegment));
  strcpy(cISASegment,"ISA");
  for (i=1; i < 17; i++)
  {
    iRecordBufferLength = strlen(cISASegment);
    cISASegment[iRecordBufferLength] =  cElementSeparator;
    strcat(cISASegment,cISA[i]);
  }
  iRecordBufferLength = strlen(cISASegment);
  cISASegment[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;

  //  Write it
  OutputFile->write(cISASegment,iRecordBufferLength);

  if (OutputFile->good())
  {
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }
}

//  Write the IEA Interchange Control Trailer Segment
int X12RecordWriter::writeIEA()
{
  char cFunctionalGroupCount[16];
  memset(cFunctionalGroupCount,0,sizeof(cFunctionalGroupCount));
  memset(RecordBuffer,0,sizeof(RecordBuffer));

  strcpy(RecordBuffer,"IEA");
  RecordBuffer[3] = cElementSeparator;

  _itoa(iFunctionalGroupCount, cFunctionalGroupCount, 10);
  strcat(RecordBuffer, cFunctionalGroupCount);
  iRecordBufferLength = strlen(RecordBuffer);
  RecordBuffer[iRecordBufferLength] = cElementSeparator;

  strcat(RecordBuffer,cInterchangeControlNumber);
  iRecordBufferLength = strlen(RecordBuffer);

  RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;

  //  Write it
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->bad())
  {
    return FAILURE;
  }

  //  Save EDI Control Document to EDI Control File Path
  writeEDIControl();
  return SUCCESS;
}

// Write the GS Functional Group Header segment
int X12RecordWriter::writeGS()
{
  //  Locals
  //  XML
  IXMLDOMElementPtr spGSInfo;
  IXMLDOMElementPtr spGroupControl;
  IXMLDOMElementPtr spTemp;
  IXMLDOMNodeListPtr spGroupsNodeList;

  //  Basic
  int iGroupControlNumber;
  int i, iLen;
  char cGSSegment[128];
  //  For GS data elements
  char cGS[9][32];
  char cGSReceiverID[32];
  char cGSType[3];
  char cControlReceiverID[32];
  char cControlGroupType[3];

  iFunctionalGroupCount++;
  iTransactionSetCount = 0;
  memset(cGSSegment,0,sizeof(cGSSegment));
  for (i=0; i< 9; i++)
  {
    memset(cGS[i],0,sizeof(cGS[i]));
  }
  memset(cGSReceiverID,0,sizeof(cGSReceiverID));
  memset(cGSType,0,sizeof(cGSType));
  memset(cControlReceiverID,0,sizeof(cControlReceiverID));
  memset(cControlGroupType,0,sizeof(cControlGroupType));


  //  GSInfo <- Call File Description Document's
  //    getElementsByTagName for "GSInformation"
  spGSInfo = 
      spFileDescription->getElementsByTagName(
                                      "GSInformation")->item[0];

  //  GSReceiverID <- Call GSInfo's getElementsByTagName for
  //  "ApplicationReceiversCode", then getAttribute for "value"
  spTemp = spGSInfo->getElementsByTagName(
                            "ApplicationReceiversCode")->item[0];
  strcpy(cGSReceiverID,_bstr_t(spTemp->getAttribute("value")));

  //  GSType <- Call GSInfo's getAttribute for
  //    "FunctionalIDCode"
  spTemp = spGSInfo->getElementsByTagName(
                            "FunctionalIDCode")->item[0];
  strcpy(cGSType,_bstr_t(spTemp->getAttribute("value")));

  //  GroupNodeList <- Call TP Elements's getElementsByTagName
  //    for "FunctionalGroup";
  spGroupsNodeList = 
      spTPElement->getElementsByTagName("FunctionalGroup");

  iLen = spGroupsNodeList->length;


  //  DO for all entries in GroupNodeList or until
  //    ControlReceiverID = GSReceiverID and
  //    ControlType = GSType
  i = 0;
  while (i < iLen)
  {
    //  ControlReceiverID <- call GroupNodeList item's
    //    getAttribute for "ReceiverID"
    memset(cControlReceiverID,0,sizeof(cControlReceiverID));
    memset(cControlGroupType,0,sizeof(cControlGroupType));
    spTemp = spGroupsNodeList->item[i];
    strcpy(cControlReceiverID,
        _bstr_t(spTemp->getAttribute("ReceiverID")));
    //  ControlType <- call GroupNodeList item's
    //    getAttribute for "FunctionalIDCode"
    strcpy(cControlGroupType,
        _bstr_t(spTemp->getAttribute("FunctionalIDCode")));

    if ( !strcmp(cControlReceiverID,cGSReceiverID) &&
         !strcmp(cControlGroupType,cGSType))
    {
      spGroupControl = spGroupsNodeList->item[i];
      break;
    }
    i++;
  } //  ENDDO

  //  IF no match found
  if ( i == iLen )
  {
    //  GroupElement <- call EDI Control's createElement method
    //    using Element name of "FunctionalGroup"
    spGroupControl =
        spEDIControl->createElement("FunctionalGroup");
    //  TP Element <- call TP Element's  appendChild to attach
    //    GroupElement
    spTPElement->appendChild(spGroupControl);
    iGroupControlNumber = 1;
    //  Call GroupElement's setAttribute for "ReceiverID",
    //    passing GSReceiverID as value to set
    spGroupControl->setAttribute("ReceiverID", cGSReceiverID);
    //  Call GroupElement's setAttribute for "FunctionalIDCode"
    //    passing GSType as value to set
    spGroupControl->setAttribute("FunctionalIDCode", cGSType);
  }
  else
  {
    //  Group Control Number <- Call TP Element's getAttribute
    //    for "LastGroupNumber"
    memset(cGroupControlNumber,0,sizeof(cGroupControlNumber));
    strcpy(cGroupControlNumber,
      _bstr_t(spGroupControl->getAttribute("LastGroupNumber")));
    iGroupControlNumber = atoi(cGroupControlNumber);
    iGroupControlNumber++;
  } //  ENDIF

  //  Call Group Element's setAttribute for "LastGroupNumber",
  //    passing Group Number as value to set
  _itoa(iGroupControlNumber, cGroupControlNumber, 10);
  spGroupControl->setAttribute("LastGroupNumber",
                            cGroupControlNumber);

  //  Get remaining values from GSInfo Element
  strcpy(cGS[1],cGSType);

  spTemp = spGSInfo->getElementsByTagName(
                            "ApplicationSendersCode")->item[0];
  strcpy(cGS[2],_bstr_t(spTemp->getAttribute("value")));

  spTemp = spGSInfo->getElementsByTagName(
                            "ApplicationReceiversCode")->item[0];
  strcpy(cGS[3],_bstr_t(spTemp->getAttribute("value")));

  spTemp = spGSInfo->getElementsByTagName(
                            "ResponsibleAgencyCode")->item[0];
  strcpy(cGS[4],_bstr_t(spTemp->getAttribute("value")));

  //  Hard code YYYYMMDD for 004010 for now
  strcpy(cGS[5],cDate8);
  strcpy(cGS[6],cTime);
  strcpy(cGS[7],cGroupControlNumber);

  spTemp = 
      spGSInfo->getElementsByTagName("VersionRelease")->item[0];
  strcpy(cGS[8],_bstr_t(spTemp->getAttribute("value")));

  //  Build the GS segment, starting with the tag
  strcpy(cGSSegment,"GS");
  for (i=1; i < 9; i++)
  {
    iRecordBufferLength = strlen(cGSSegment);
    cGSSegment[iRecordBufferLength] =  cElementSeparator;
    strcat(cGSSegment,cGS[i]);
  }
  iRecordBufferLength = strlen(cGSSegment);
  cGSSegment[iRecordBufferLength] =  cRecordTerminator1;
  iRecordBufferLength++;

  //  Write it
  OutputFile->write(cGSSegment,iRecordBufferLength);

  if (OutputFile->good())
  {
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }
}

//  Write the GE Functional Group Trailer Segment
int X12RecordWriter::writeGE()
{
  memset(cTransactionSetCount,0,sizeof(cTransactionSetCount));
  memset(RecordBuffer,0,sizeof(RecordBuffer));
  strcpy(RecordBuffer,"GE");
  iRecordBufferLength = 2;
  RecordBuffer[iRecordBufferLength] = cElementSeparator;
  _itoa(iTransactionSetCount, cTransactionSetCount, 10);
  strcat(RecordBuffer, cTransactionSetCount);
  iRecordBufferLength = strlen(RecordBuffer);
  RecordBuffer[iRecordBufferLength] = cElementSeparator;
  strcat(RecordBuffer,cGroupControlNumber);
  iRecordBufferLength = strlen(RecordBuffer);
  RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;
  iTransactionSetCount = 0;
  //  Write it
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->bad())
  {
    return FAILURE;
  }
  return SUCCESS;
}

// Write the ST Transaction Set Header segment
int X12RecordWriter::writeST()
{
  //  Locals
  //  XML
  IXMLDOMElementPtr spSTInfo;
  IXMLDOMElementPtr spTemp;

  //  Basic
  int i, iLen, iPad;
  char cSTSegment[32];

  //  For ST data elements
  char cST[4][32];
  char cSTType[4];
  
  memset(cSTSegment,0,sizeof(cSTSegment));
  memset(cSTType,0,sizeof(cSTType));
  for (i=0; i < 4; i++)
  {
    memset(cST[i],0,sizeof(cST[i]));
  }
  memset(cTransactionSetCount,0,sizeof(cTransactionSetCount));

  iTransactionSetCount++;
  _itoa(iTransactionSetCount, cTransactionSetCount, 10);

  //  Pad Transaction Set Count to four digits with leading
  //    zeroes if required
  iLen = strlen(cTransactionSetCount);
  if (iLen < 4)
  {
    iPad = 4 - iLen;
    strcpy(&(cTransactionSetCount[iPad]),cTransactionSetCount);
    memset(cTransactionSetCount,'0',iPad);
  }

  //  STInfo <- Call File Description Document's
  //    getElementsByTagName for "STInformation"
  spSTInfo = 
      spFileDescription->getElementsByTagName(
                                    "STInformation")->item[0];

  //  Set values
  spTemp = spSTInfo->getElementsByTagName(
                                "TransactionSetIDCode")->item[0];
  strcpy(cST[1], _bstr_t(spTemp->getAttribute("value")));

  strcpy(cST[2],cTransactionSetCount);

  //  Include support for ST03 Implementation Convention
  //    Reference - 004030 and later
  spTemp = spSTInfo->getElementsByTagName(
                  "ImplementationConventionReference")->item[0];
  if (spTemp != NULL)
  {
    strcpy(cST[3],_bstr_t(spTemp->getAttribute("value")));
  }

  //  Build the ST segment, starting with the tag
  strcpy(cSTSegment,"ST");
  for (i=1; i < 3; i++)
  {
    iRecordBufferLength = strlen(cSTSegment);
    if (strlen(cST[i]) != 0)
    {
      cSTSegment[iRecordBufferLength] = cElementSeparator;
      strcat(cSTSegment,cST[i]);
    }
  }
  //  Append the segment terminator
  iRecordBufferLength = strlen(cSTSegment);
  cSTSegment[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;

  iSegmentCount = 1;

  //  Write it
  OutputFile->write(cSTSegment,iRecordBufferLength);
  if (OutputFile->good())
  {
    return SUCCESS;
  }
  else
  {
    return FAILURE;
  }
}

//  Write the SE Transaction Set Trailer Segment
int X12RecordWriter::writeSE()
{
  char cSegmentCount[16];
  memset(cSegmentCount,0,sizeof(cSegmentCount));

  iSegmentCount++;
  memset(RecordBuffer,0,sizeof(RecordBuffer));
  strcpy(RecordBuffer,"SE");
  iRecordBufferLength = 2;
  RecordBuffer[iRecordBufferLength] = cElementSeparator;
  _itoa(iSegmentCount, cSegmentCount, 10);
  strcat(RecordBuffer, cSegmentCount);
  iRecordBufferLength = strlen(RecordBuffer);
  RecordBuffer[iRecordBufferLength] = cElementSeparator;
  strcat(RecordBuffer,cTransactionSetCount);
  iRecordBufferLength = strlen(RecordBuffer);
  RecordBuffer[iRecordBufferLength] = cRecordTerminator1;
  iRecordBufferLength++;
  iSegmentCount = 0;
  //  Write it
  OutputFile->write(RecordBuffer,iRecordBufferLength);

  if (OutputFile->bad())
  {
    return FAILURE;
  }
  return SUCCESS;
}
