/*  X12SourceConverter.cpp

Drives conversion from an X12 interchange to XML

From Chapter 9 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include <direct.h>
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"


//  Program specific header files
#include "EDIRecordReader.h"
#include "X12RecordReader.h"
#include "X12SourceConverter.h"
#include <strstrea.h>


//  Default Constructor
X12SourceConverter::X12SourceConverter(bool boValidateIn, 
    const char * const cBaseOutputDirectoryIn,
    const char * const cFileDescriptionDocumentPath)
    : SourceConverter(boValidateIn, cBaseOutputDirectoryIn)
{
#include "BlasterExceptions.h"
  //  Locals

  int iReturn;

  //  FDD Path <- From passed path name
  memset(cFDDPath,0,sizeof(cFDDPath));
  strcpy(cFDDPath,cFileDescriptionDocumentPath);
  memset(cSavedFDDName,0,sizeof(cSavedFDDName));

  //  Create FA_Out diretory
  memset(cFAOutputDirectory,0,sizeof(cFAOutputDirectory));
  strcpy(cFAOutputDirectory,cBaseOutputDirectory);
  strcat(cFAOutputDirectory,"FA_Out\\");
    
  iReturn = _mkdir(cFAOutputDirectory);
  if (iReturn != SUCCESS)
  {
    cerr << IO_DIRECTORY_CREATE_ERROR_MESSAGE
         << cFAOutputDirectory << endl;
    throw cConstructorError;
  }

  //  Create FlatRecordReader object, passing:
  //    File Description Document DOM
  X12Reader = new X12RecordReader(NULL);
  Reader = X12Reader;
}


//  Destructor
X12SourceConverter::~X12SourceConverter() {};


//  Process a single file
int X12SourceConverter::processFile(
    const char * const cInputFileName)
{
  //  Locals to method

  //  COM
  
  //  XML

  //  File input and output streams
  ifstream InputFile;

  //  Basic types
  int iRecordLength, iReturn;

  //  Initializations

  //  Open input file
  InputFile.open(cInputFileName, ios::in | ios::nocreate);
  if ( ! InputFile )
  {
    cerr << INPUT_FILE_OPEN_ERROR << cInputFileName << endl;
    return FAILURE;
  }

    //  C++ Only - set mode to binary
  InputFile.setmode(filebuf::binary);


  //  Call RecordReader's setInputStream method
  X12Reader->setInputStream(&InputFile);

  X12Reader->parseISA();
  X12Reader->logISA();
  
  //  Call RecordReader's readRecord method
  iRecordLength = X12Reader->readRecordVariableLength();

  //  IF RecordLength = 0
  if (iRecordLength == 0)
  {
    cerr << INVALID_INTERCHANGE << "GS" << endl;
    return FAILURE;
  }

  //  Segment Tag <- Call FlatRecordReader's getRecordType method
  memset(cTag,0,sizeof(cTag));
  iReturn = X12Reader->getRecordType(cTag);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  DO while Tag = GS to process all functional groups
  while (!strcmp(cTag,"GS"))
  {
    iReturn = X12Reader->parseGS();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    iReturn = X12Reader->logGS();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
 
    //  Get GS02
    memset(cGroupSenderID,0,sizeof(cGroupSenderID));
    iReturn = X12Reader->getControlSegmentElement("GS",2,
        cGroupSenderID);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
      
    //  Get GS01
    memset(cGroupID,0,sizeof(cGroupID));
    iReturn = X12Reader->getControlSegmentElement("GS",1,
        cGroupID);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Get GS08
    memset(cX12Version,0,sizeof(cX12Version));
    iReturn = X12Reader->getControlSegmentElement("GS",8,
        cX12Version);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Output Directory Path <- Base Directory +
    //    GroupSenderID + GroupID + directory separator
    memset(cOutputDirectoryPath,0,sizeof(cOutputDirectoryPath));
    strcpy(cOutputDirectoryPath,cBaseOutputDirectory);
    strcat(cOutputDirectoryPath,cGroupSenderID);
    strcat(cOutputDirectoryPath,cGroupID);
    strcat(cOutputDirectoryPath,"\\");

    //  Create the directory for the group
    iReturn = _mkdir(cOutputDirectoryPath);
    if (iReturn != SUCCESS)
    {
      cerr << IO_DIRECTORY_CREATE_ERROR_MESSAGE
           << cOutputDirectoryPath<< endl;
      return FAILURE;
    }

    iRecordLength = X12Reader->readRecordVariableLength();
    if (iRecordLength == 0)
    {
      cerr << INVALID_INTERCHANGE << "ST" << endl;
      return FAILURE;
    }

    //  Segment Tag <- Call FlatRecordReader's getRecordID method
    memset(cTag,0,sizeof(cTag));
    iReturn = X12Reader->getRecordType(cTag);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    iReceivedTransactionSets = 0;

    //  DO while (Tag = ST) to process all transaction sets
    while (!strcmp(cTag,"ST"))
    {
      iReceivedTransactionSets++;
      iReturn = processDocument();
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    } //  ENDDO for transaction sets

    //  IF Tag != GE
    if (strcmp(cTag,"GE"))
    {
      cerr << INVALID_INTERCHANGE << "GE" << endl;
      return FAILURE;
    }
    X12Reader->parseGE();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    X12Reader->logGE();
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  Included Transaction Sets <- GE01
    memset(cIncludedTransactionSets,0,
        sizeof(cIncludedTransactionSets));
    iReturn = X12Reader->getControlSegmentElement("GE",1,
        cIncludedTransactionSets);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }

    //  IF GSID != "FA"
    //  Call writeFunctionalAck
    if (strcmp(cGroupID,"FA"))
    {
      writeFunctionalAck();
    }

    //  RecordLength <- RecordReader readRecordVariableLength
    iRecordLength = X12Reader->readRecordVariableLength();

    //  IF RecordLength != 0
    if (iRecordLength != 0)
    {
      //  Tag <- Call X12RecordReader's getRecordType
      memset(cTag,0,sizeof(cTag));
      iReturn = X12Reader->getRecordType(cTag);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }
    }
    else
    {
      cerr << INVALID_INTERCHANGE << "GS or IEA" << endl;
      return FAILURE;
    }
  } //  ENDDO for processing functional groups

  //  IF (Tag != IEA)
  if (strcmp(cTag,"IEA"))
  {
    cerr << INVALID_INTERCHANGE << "IEA" << endl;
    return FAILURE;
  }

  iReturn = X12Reader->parseIEA();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  iReturn = X12Reader->logIEA();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  // Close input file
  InputFile.close();

  return SUCCESS;
}

//  Process a group of records
int X12SourceConverter::processDocument()
{
  //  Locals
  HRESULT hResult;

  //  XML
  IXMLDOMElementPtr spRoot;
  IXMLDOMElementPtr spSegmentGrammar;
  IXMLDOMNodePtr spChildNode;
  IXMLDOMElementPtr spEleTemp;
  IXMLDOMNodeListPtr spNLTemp;
  IXMLDOMProcessingInstructionPtr spProlog;
  IXMLDOMAttributePtr spSchemaLocationAttribute;

  //  Basic
  char cGrammarElementName[NMTOKEN_SIZE];
  char cSegmentElementName[NMTOKEN_SIZE];
  char cGrammarTag[CELL_BUFFER_SIZE];
  char cTransactionSetID[4];
  char cTransactionSetControlNumber[10];
  char cFDDName[MAX_PATH_LENGTH];
  char cOutputFilePath[MAX_PATH_LENGTH];

  int iRecordLength, iLen, iReturn;

  //  Initializations
  memset(cGrammarElementName,'\0',sizeof(cGrammarElementName));
  memset(cSegmentElementName,'\0',sizeof(cSegmentElementName));
  memset(cFDDName,'\0',sizeof(cFDDName));

  spRoot = NULL;

  iReturn = X12Reader->parseST();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }
  iReturn = X12Reader->logST();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  TrasactionSetID <- From ST01
  iReturn = X12Reader->getControlSegmentElement("ST",1,
      cTransactionSetID);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  FDDName <- FDD Path + GroupSenderID + '-' + GroupID +
  //    '-' + TransactionSetID + ".xml"
  strcpy(cFDDName,cFDDPath);
  strcat(cFDDName,cGroupSenderID);
  strcat(cFDDName,"-");
  strcat(cFDDName,cGroupID);
  strcat(cFDDName,"-");
  strcat(cFDDName,cTransactionSetID);
  strcat(cFDDName,"-");
  strcat(cFDDName,cX12Version);
  strcat(cFDDName,".xml");


  //  IF (FDDName != Saved FDD Name)
  if (strcmp(cFDDName,cSavedFDDName))
  {
    //  Call loadFileDescriptionDocument, passing FDDName
    iReturn = loadFileDescriptionDocument(cFDDName);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
    
    memset(cSavedFDDName,0,sizeof(cSavedFDDName));
    strcpy(cSavedFDDName,cFDDName);
    //  Schema Location URL <- Call File Description Document's
    //    getElementByTagName on "SchemaLocationURL", then
    //    getAttribute on "value"
    //  NOTE - This is an optional Element, so we may not
    //  get anything
    memset(cSchemaLocationURL,0,sizeof(cSchemaLocationURL));
    spNLTemp = spFileDescription->getElementsByTagName(
        "SchemaLocationURL");
    if (spNLTemp->length != 0)
    {
      spEleTemp = spNLTemp->item[0];
      strcpy(cSchemaLocationURL,
        _bstr_t(spEleTemp->getAttribute("value")));
    }

    //  Throw exception if no schema and validation requested
    if ( boValidate && (strlen(cSchemaLocationURL) == 0) )
    {
      cerr << VALIDATION_WITHOUT_SCHEMA << endl;
      return FAILURE;
    }
  }

  //  Create new Output DOM Document
  hResult = spDocOutput.CreateInstance(__uuidof(DOMDocument40));

  if FAILED(hResult)
  {
    cerr << CREATE_INSTANCE_ERROR << endl;
    return FAILURE;
  }

  //  Create the header processing instruction
  //    and append it to the document node
  spProlog = spDocOutput->createProcessingInstruction("xml",
    XML_PROLOG_TEXT);
  spDocOutput->appendChild(spProlog);

  //  Create Root Element, using Root Element Name from Grammar,
  //    and append to Output DOM Document
  spRoot = spDocOutput->createElement(cRootElementName);
  spDocOutput->appendChild(spRoot);

   //  IF Schema Location URL is not NULL
  if (strlen(cSchemaLocationURL) > 0)
  {
    //  Create noNameSpaceSchemaLocation Attribute and
    //    append to Root Element
    _variant_t varType((short)NODE_ATTRIBUTE);
    spSchemaLocationAttribute = spDocOutput->createNode(
      varType,"xsi:noNamespaceSchemaLocation",
      "http://www.w3.org/2001/XMLSchema-instance");

    spRoot->setAttributeNode(spSchemaLocationAttribute);
    // We can finally set it now

    spRoot->setAttribute("xsi:noNamespaceSchemaLocation",
      cSchemaLocationURL);
  }  //    ENDIF

  //  Call RecordReader's setOutputDocument method for new
  //    document
  X12Reader->setOutputDocument(spDocOutput);

  //  TransactionSetControlNum <- ST02
  memset(cTransactionSetControlNumber,0,
      sizeof(cTransactionSetControlNumber));
  iReturn = X12Reader->getControlSegmentElement("ST",2,
      cTransactionSetControlNumber);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  // Trim transaction set control number
  iLen = strlen(cTransactionSetControlNumber);
  while(cTransactionSetControlNumber[iLen - 1] == ' ')
  {
    cTransactionSetControlNumber[iLen - 1] = 0;
    iLen--;
  }
  int iPos = 0;
  while(cTransactionSetControlNumber[iPos] == ' ')
  {
    iPos++;
  }
  if(iPos >= 1)
  {
    strcpy(cTransactionSetControlNumber,
      &(cTransactionSetControlNumber[iPos]));
    memset(&(cTransactionSetControlNumber[iLen - iPos]),0,iPos);
  }  

  //  Output File Path <- Output directory path + Root Element
  //    Name + TransactionSetControlNum + ".xml"
  memset(cOutputFilePath,0,sizeof(cOutputFilePath));
  strcpy(cOutputFilePath, cOutputDirectoryPath);
  strcat(cOutputFilePath, cRootElementName);
  strcat(cOutputFilePath, cTransactionSetControlNumber);
  strcat(cOutputFilePath, ".xml");

  //  Display document name
  cout << "    Saving to: " << cOutputFilePath << endl;

  iRecordLength = X12Reader->readRecordVariableLength();

  //  IF RecordLength = 0
  if (iRecordLength == 0)
  {
    cout << INVALID_INTERCHANGE
         << "No segments after ST" << endl;
    return FAILURE;
  }

  //  Tag <- Call X12RecordReader's getRecordType
  memset(cTag,0,sizeof(cTag));
  iReturn = X12Reader->getRecordType(cTag);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  ChildNode <- Get Grammar Element's firstChild
  spChildNode = spGrammar->firstChild;

  //  DO while ChildNode is not an Element Node
  while (spChildNode->nodeType != NODE_ELEMENT)
  {
    spChildNode = spChildNode->nextSibling;
  }

  //  Segment Grammar Element <- ChildNode
  spSegmentGrammar = spChildNode;

  //  Grammar Record Tag <- Segment Grammer getAttribute
  //    TagValue
  strcpy(cGrammarTag,
    _bstr_t(spSegmentGrammar->getAttribute("TagValue")));

  //  DO until (Tag = SE) at end of transaction set
  while (strcmp(cTag,"SE"))
  {
    //  DO until Grammar Record Tag = Segment Tag
    while (strcmp(cTag,cGrammarTag))
    {
      //  Segment Grammar Element <- Get next Segment Element
      //    from Grammar
      spChildNode = spChildNode->nextSibling;

      //  IF Segment Grammar Element is NULL
      if (spChildNode == NULL)
      {
        // This Segment is not part of the document
        cerr << NOT_IN_DOCUMENT_GRAMMAR << cTag << endl;
        return FAILURE;
      }  // ENDIF

      if (spChildNode->nodeType != NODE_ELEMENT)
      {
        continue;
      }

      spSegmentGrammar = spChildNode;

      //  Grammar Record Tag <- Record Grammer getAttribute
      //    TagValue
      memset(cGrammarTag,'\0',sizeof(cGrammarTag));
      strcpy(cGrammarTag,
        _bstr_t(spSegmentGrammar->getAttribute("TagValue")));
    } // ENDDO until Grammar Tag = Tag

    //  Grammar Element Name <- Record Grammar getNodeName
    memset(cGrammarElementName,'\0',sizeof(cGrammarElementName));
    strcpy(cGrammarElementName, spSegmentGrammar->nodeName);

    //  IF Grammar Element Name = "GroupDescription"
    if (!(strcmp(cGrammarElementName,"GroupDescription")))
    {
      //  Call processGroup
      iReturn = processGroup(spRoot, spSegmentGrammar,
        iRecordLength);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

    }
    else
    {
      //  Call RecordReader's parseRecord
      iReturn = X12Reader->parseRecord(spSegmentGrammar);
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  Call RecordReader's toXML
      iReturn = X12Reader->toXMLType();
      if (iReturn != SUCCESS)
      {
        return iReturn;
      }

      //  Call writeRecord
      X12Reader->writeRecord(spRoot, spSegmentGrammar);

      //  Input Record <- readRecord
      iRecordLength = X12Reader->readRecordVariableLength();
    } //  ENDIF

    //  IF RecordLength = 0
    if (iRecordLength == 0)
    {
      cerr << INVALID_INTERCHANGE << "No SE segment" << endl;
      return FAILURE;
    }
    memset(cTag,0,sizeof(cTag));
    iReturn = X12Reader->getRecordType(cTag);
    if (iReturn != SUCCESS)
    {
      return iReturn;
    }
  } //  ENDDO until (Tag = SE) at end of transaction set

  X12Reader->parseSE();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }
  X12Reader->logSE();
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  //  Call saveDocument
  saveDocument(spDocOutput, cOutputFilePath, boValidate);

  iRecordLength = X12Reader->readRecordVariableLength();
  //  IF RecordLength = 0
  if (iRecordLength == 0)
  {
    cerr << INVALID_INTERCHANGE
         << "No GS or GE segment following SE" << endl;
    return FAILURE;
  }

  memset(cTag,0,sizeof(cTag));
  iReturn = X12Reader->getRecordType(cTag);
  if (iReturn != SUCCESS)
  {
    return iReturn;
  }

  return SUCCESS;
}

//  Write the XML representation of a 997 functional
//    acknowledgement transaction set
int X12SourceConverter::writeFunctionalAck()
{
  //  Locals
  HRESULT hResult;

  //  XML
  IXMLDOMDocumentPtr spDocFA;
  IXMLDOMProcessingInstructionPtr spProlog;
  IXMLDOMElementPtr spFARoot;
  IXMLDOMElementPtr spSegment;
  IXMLDOMElementPtr spDataElement;
  IXMLDOMTextPtr spText;

  //  Basic
  char cFAOutputFileName[MAX_PATH_LENGTH];
  char cGSControlNum[16];
  char cReceivedTransactionSets[8];

  //  Output File Name <- FA Output Path + GS Sender ID + "-"
  //    + Functional Group ID Code + ".xml"
  memset(cFAOutputFileName,0,sizeof(cFAOutputFileName));
  strcpy(cFAOutputFileName,cFAOutputDirectory);
  strcat(cFAOutputFileName,cGroupSenderID);
  strcat(cFAOutputFileName,"-");
  strcat(cFAOutputFileName,cGroupID);
  strcat(cFAOutputFileName,".xml");

  //  Create new Output DOM Document
  hResult = spDocFA.CreateInstance(__uuidof(DOMDocument40));

  if FAILED(hResult)
  {
    cerr << CREATE_INSTANCE_ERROR << endl;
    return FAILURE;
  }

  //  Create the header processing instruction
  //    and append it to the document node
  spProlog = spDocOutput->createProcessingInstruction("xml",
    XML_PROLOG_TEXT);
  spDocFA->appendChild(spProlog);

  //  Create Root Element "TS997" and append to FA document
  spFARoot = spDocFA->createElement("TS997");
  spDocFA->appendChild(spFARoot);

  //  Create AK1 Element and append to Root
  spSegment = spDocFA->createElement("AK1");
  spFARoot->appendChild(spSegment);

  //  Create AK101 Element and append to AK1
  spDataElement = spDocFA->createElement("AK101");
  spSegment->appendChild(spDataElement);

  //  Create Text Node with Functional Group ID Code
  //    and append to AK101
  spText = spDocFA->createTextNode(cGroupID);
  spDataElement->appendChild(spText);

  //  Create AK102 Element and append to AK1
  spDataElement = spDocFA->createElement("AK102");
  spSegment->appendChild(spDataElement);

  //  Create Text Node with GS Control Number and append to
  //    AK102
  memset(cGSControlNum,0,sizeof(cGSControlNum));
  X12Reader->getControlSegmentElement("GS",6,cGSControlNum);
  spText = spDocFA->createTextNode(cGSControlNum);
  spDataElement->appendChild(spText);

  //  Create AK9 Element and append to Root
  spSegment = spDocFA->createElement("AK9");
  spFARoot->appendChild(spSegment);

  //  Create AK901 Element and append to AK9
  spDataElement = spDocFA->createElement("AK901");
  spSegment->appendChild(spDataElement);

  //  Create Text Node with "A" and append to AK101
  spText = spDocFA->createTextNode("A");
  spDataElement->appendChild(spText);

  //  Create AK902 Element and append to AK9
  spDataElement = spDocFA->createElement("AK902");
  spSegment->appendChild(spDataElement);

  // AK902 is count of included transaction sets from GE01
  //  Create Text Node with Included Transaction Sets
  //  and append to AK902
  spText = spDocFA->createTextNode(cIncludedTransactionSets);
  spDataElement->appendChild(spText);

  // AK903 is count of transaction sets actually received
  //  Create AK903 Element and append to AK9
  spDataElement = spDocFA->createElement("AK903");
  spSegment->appendChild(spDataElement);

  //  Create Text Node with Received Transaction Sets
  //    and append to AK903
  memset(cReceivedTransactionSets,0,
      sizeof(cReceivedTransactionSets));
  _itoa(iReceivedTransactionSets,cReceivedTransactionSets,10);
  spText = spDocFA->createTextNode(cReceivedTransactionSets);
  spDataElement->appendChild(spText);

  // AK904 is count of transaction sets accepted - We're going
  //   to say we accepted all of them that we received
  //  Create AK904 Element and append to AK9
  spDataElement = spDocFA->createElement("AK904");
  spSegment->appendChild(spDataElement);

  //  Create Text Node with Received Transaction Sets
  //    and append to AK904
  spText = spDocFA->createTextNode(cReceivedTransactionSets);
  spDataElement->appendChild(spText);

  //  Save FA Document
  saveDocument(spDocFA, cFAOutputFileName, false);
  return SUCCESS;
}



