/* X12ToXML.cpp

This program converts an X12 interchange into one or more XML
instance documents

Program developed in Chapter 9
Using XML with Legacy Business Applications

Source code is Copyright (c) 2002 Addison-Wesley Professional
Author:
Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License,
Version 2, as published by the Free Software Foundation.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"
#include "BlasterPrototypes.h"

//  Program specific header files
#include "EDIRecordReader.h"
#include "X12RecordReader.h"
#include "X12SourceConverter.h"


int main(int argc, char* argv[])
{
#include "BlasterExceptions.h"
  // Local variables

  //  COM & MS
  HRESULT hResult;

  //  Project Classes


  //  Local variables

  //  File input and output streams
  ifstream InputStream;

  //  Basic types
  char * cInputFileName;
  char cOuputDirectory[MAX_PATH_LENGTH];
  char * cFileDescriptionDocument;
  char cUsageMessage[256];
  char cHelpMessage[1024];
  bool boValidate;
  int iExitStatus;
  int iDocumentsProcessed;
  int i;
  int iReturn;

  //  Initializations
  memset(cUsageMessage,0,sizeof(cUsageMessage));
  strcpy(cUsageMessage,X12_TO_XML_USAGE);

  memset(cHelpMessage,0,sizeof(cHelpMessage));
  strcpy(cHelpMessage,X12_TO_XML_USAGE);
  strcat(cHelpMessage,X12_TO_XML_HELP);
  strcat(cHelpMessage,HELP_INFO_1);
  strcat(cHelpMessage,"X12ToXML");
  strcat(cHelpMessage,HELP_INFO_2);

  iExitStatus = SUCCESS;
  boValidate = false;
  iDocumentsProcessed = 0;

  memset(cOuputDirectory,0,MAX_PATH_LENGTH);


  // Parse input arguments from command line
  // IF help option is specified
  //     display help message and exit
  // ENDIF
  if ( ( argc == 2 ) && (strcmp(argv[1],"-h") == 0 ) )
  {
    cout << cHelpMessage << endl << endl;
    return SUCCESS;
  }
    
  if ((argc < 4) || (argc > 7) )
  {
    cout << cUsageMessage << endl;
    return FAILURE;
  }

  // Check for validation and help message
  for (i = 4; i < argc; i++)
  {
    if (argv[i][0] == '-')
    {
      switch (argv[i][1])
      {
        case 'h':
          cout << cHelpMessage << endl;
          return SUCCESS;
        case 'v':
          boValidate = true;
          break;
        default:
          cout << cUsageMessage << endl;
          return FAILURE;
      }
    }
    else
    {
      cout << cUsageMessage << endl;
      return FAILURE;
    }
  }

  if (strlen(argv[2]) > (MAX_PATH_LENGTH - 2))
  {
    cerr << BUFFER_OVERFLOW_IN << "main" << endl;
    return FAILURE;
  }

  cInputFileName = argv[1];
  strcpy(cOuputDirectory,argv[2]);
  cFileDescriptionDocument = argv[3];

  //  Set up the COM environment, since MSXML is a
  //  COM Component
  hResult = CoInitialize(NULL);
  if (FAILED(hResult))
  {
    cerr << COM_INIT_ERROR << endl;
    return FAILURE;
  }

  //  We'll do our main processing within a try block
  //  We only expect to throw COM, MSXML parse
  //  and I/O exceptions
  try
  {
    //  Create new X12SourceConverter object, passing:
    //    Output Directory
    //    File Description XML Document
    //    Validation option
    X12SourceConverter Converter(boValidate, cOuputDirectory, 
        cFileDescriptionDocument);

    // Call CSVSourceConverter processFile method, passing
    //    the input file name
    iReturn = Converter.processFile(cInputFileName);
    if (iReturn != SUCCESS)
    {
      throw cProcessFileError;
    }
  } //  End of try block
    
 
  //  General purpose catch block
  catch (char cException)
  {
    switch (cException)
    {
      case cCreateInstanceError:
      case cConstructorError:
      case cXMLValidationError:
      case cProcessFileError:
        break;
      default:
        cerr << UNEXPECTED_EXCEPTION << cException
             << endl;
    }
    iExitStatus = FAILURE;
  }
    
  //  Do some simple error handling for COM exceptions
  catch (_com_error &e)
  {
    handleCOMError(e);
    iExitStatus = FAILURE;
  }

  // Close COM Environment
  CoUninitialize();

  // Finish up
  if (iExitStatus == SUCCESS)
  {
    cout << TERMINATION_NORMAL << endl;
  }
  else
  {
    cout << ABEND << endl;
  }

  return iExitStatus;
}
