/* DataCellDateMMsDDsYYYY.cpp

For handling of Date data type in MM/DD/YYYY format
fields in non-XML files

From Chapter 7 of
Using XML with Legacy Business Applications
Copyright (c) 2002 Addison-Wesley Professional

Author: Michael C. Rawlins
Rawlins EC Consulting
PMB 29
14 Canyon Creek Village
Richardson, TX 75080-1602
mike@rawlinsecconsulting.com
972-783-9573

This file is one of several utilities developed for the book.

All utilities developed for the book are free software; you can
redistribute and/or modify them under the terms of the GNU
General Public License, Version 2, as published by the Free
Software Foundation.

These utilities are distributed in the hope that they will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the book; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA

The license is also available on the World Wide Web at
http://www.gnu.org/licenses/gpl.html

*/

//  Common header files
#include "BlasterIncludes.h"

//  Constructor with field number and grammar Element
DataCellDateMMsDDsYYYY::DataCellDateMMsDDsYYYY(int iFieldNumberIn,
    const IXMLDOMElementPtr spGrammarIn) :
  DataCell(iFieldNumberIn, spGrammarIn) { }

//  Destructor
DataCellDateMMsDDsYYYY::~DataCellDateMMsDDsYYYY() {};


//  Convert from ISO YYYY-MM-DD to MM/DD/YYYY
int DataCellDateMMsDDsYYYY::fromXML()
{
  //  Locals
  char cTarget[11];
  memset(cTarget,0,sizeof(cTarget));

  //  IF length of CellBuffer != 10
  if (strlen(cCellBuffer) != 10)
  {
    //  Return error
    cerr << BAD_DATE << cCellBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  TempBuffer <- CellBuffer characters 5 and 6
  //  TempBuffer <- Append forward slash
  //  TempBuffer <- Append CellBuffer characters 8 and 9
  //  TempBuffer <- Append forward slash
  //  TempBuffer <- Append CellBuffer characters 0 through 3
  strncpy(cTarget,&(cCellBuffer[5]),2);
  cTarget[2] = '/';
  strncat(cTarget,&(cCellBuffer[8]),2);
  cTarget[5] = '/';
  strncat(cTarget,cCellBuffer,4);

  //  CellBuffer <- TempBuffer
  putField(cTarget,10);

  return SUCCESS;
}



//  Prepare output for legacy format
//  This method developed in Chapter 8
int DataCellDateMMsDDsYYYY::prepareOutput()
{
  char cLength[5];
  int iLength = 0;
  int iMinLength = 0;
  _variant_t vLength;

  memset(cLength,0,sizeof(cLength));

  //  Length <- Call Grammar Element's getAttribute for
  //    "Length"
  //  We may have Length or MinLength - With MSXML we get
  //  a variant and have to check if it is null
  vLength = spGrammar->getAttribute("Length");
  if (vLength.vt != VT_NULL)
  {
    strcpy(cLength,_bstr_t(vLength));
    if (strlen(cLength) > 0)
    {
      iLength = atoi(cLength);
      iMinLength = iLength;
    }
  }
  else
  {
    strcpy(cLength,
      _bstr_t(spGrammar->getAttribute("MinLength")));
    iMinLength = atoi(cLength);
  }

  //  Call fillField, passing L for left justification, and
  //    Length
  fillField('L', iMinLength);

  //  Return success
  return SUCCESS;
}


//  Convert from MM/DD/YYYY to ISO YYYY-MM-DD
int DataCellDateMMsDDsYYYY::toXML()
{
#define MonthState 0
#define DayState 1
#define YearState 2

  //  Locals
  char cTarget[11];
  char cMonth[11];
  char cDay[11];
  char cYear[11];
  char cTempChar;

  //  State <- Month
  int iParseState = MonthState;
  int i;


  memset(cTarget,0,sizeof(cTarget));
  memset(cMonth,0,sizeof(cMonth));
  memset(cDay,0,sizeof(cDay));
  memset(cYear,0,sizeof(cYear));

  //  CellBuffer < trim leading and trailing white space from
  //    CellBuffer
  trim();

  //  IF length of CellBuffer != 10
  if ( iBufferLength > 10)
  {
    //  Return error
    cerr << BAD_DATE << cCellBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  TempChar <- First character in CellBuffer
  //  DO until end of CellBuffer
  //  TempChar <- Next character in CellBuffer
  for (i = 0; i < iBufferLength; i++)
  {
    cTempChar = cCellBuffer[i];
    //  IF TempChar = Slash
    if ( cTempChar == '/' )
    {
      //  Increment State
      iParseState++;
    }
    else  //  ELSE
    {
      //  DO CASE of State
      switch (iParseState)
      {
        //  Month:
        case MonthState:
          //  Append TempChar to Month
          cMonth[strlen(cMonth)] = cTempChar;
          break;  //  BREAK
        //  Day:
        case DayState:
          //  Append TempChar to Day
          cDay[strlen(cDay)] = cTempChar;
          break;  //  BREAK
        //  Year:
        case YearState:
          //  Append TempChar to Year
          cYear[strlen(cYear)] = cTempChar;
          break;  //  BREAK
        //  other:
        default:
          //  Return Error
          cerr << BAD_DATE << cCellBuffer << endl;
          return FAILURE;
      } //  ENDDO
    } //  ENDIF
  } //  ENDDO

  //  IF length of Month is 1
  if (strlen(cMonth) == 1)
  {
    //  Month <- "0" + Month
    cMonth[1] = cMonth[0];
    cMonth[0] = '0';
  } //  ENDIF

  //  If Length of Day is 1
  if (strlen(cDay) == 1)
  {
    //  Day <- "0" + Day
    cDay[1] = cDay[0];
    cDay[0] = '0';
  } //  ENDIF

  //  IF (Length of Month != 2) OR (Length of Day != 2) OR
  //    (Length of Year != 4)
  if ( (strlen(cMonth) != 2) || (strlen(cDay) != 2) ||
    (strlen(cYear) != 4) )
  {
    //  Return Error
    cerr << BAD_DATE << cCellBuffer << endl;
    return FAILURE;
  } //  ENDIF

  //  CellBuffer = Year + dash + Month + dash + Day
  strcpy(cTarget,cYear);
  strcat(cTarget,"-");
  strcat(cTarget,cMonth);
  strcat(cTarget,"-");
  strcat(cTarget,cDay);

  putField(cTarget,10);

  return SUCCESS;
}
